--[[
	Chris Hooks V2
	Q. What does this script do?
	A. This script makes creating characters and gamemodes just a bit easier.
]]--
--[[
	DOCUMENTATION
	# SKIN HOOKS
	
	* SkinEnabled(p)
	* SkinDisabled(p)
	* PreThinkFrame(p)
	* PostThinkFrame(p)
	* PlayerThink(p)
	* ShouldDamage(p,...)
	* PlayerDamage(p,...)
	* PlayerAttack(p,...)
	* PlayerCmd(p,...)
	
	-- All of the action hooks (i aint writing them down again)
	
	# GAMETYPE HOOKS
	
	* GametypeDisabled()
	* PreThinkFrame(p) -- NOTE: Make sure to check if p is nil, it runs once without the player.
	* PostThinkFrame(p) -- Same warning as above
	* ThinkFrame()
	* MapLoad(mapid)
	* MapChange(mapid)
	* IntermissionThinker(...)
	* VoteThinker(...)
	* PlayerQuit(...)
	* TeamSwitch(...)
	
]]--

if(rawget(_G,"chrishooks_srb2_api"))then return end

local hook = {
	s = {}, -- Skin Hooks
	g = {}, -- Gamemode Hooks
	activePlayers = 0
}
local lastGametype = nil

addHook("NetVars",function(n)
	lastGametype = n($)
end)

-- Functions

local function enableAllHud()
	hud.enable("stagetitle")
	hud.enable("textspectator")
	hud.enable("crosshair")
	hud.enable("score")
	hud.enable("time")
	hud.enable("gametypeinfo")
	hud.enable("minimap")
	hud.enable("item")
	hud.enable("position")
	--hud.enable("names") -- ????
	hud.enable("check")
	hud.enable("minirankings")
	hud.enable("battlerankingsbumpers")
	hud.enable("wanted")
	hud.enable("speedometer")
	hud.enable("freeplay")
	hud.enable("rankings")
	
	hud.enable("intermissiontally")
	hud.enable("intermissionmessages")
end


-- Skin Hooks

local function doSkinHook(skin,h,...)
	local s = skins[skin].name
	if(not hook.s[s])then return end
	if(not hook.s[s][h])then return nil end
	for i,f in ipairs(hook.s[s][h]) do
		local val = {f(...)}
		if(#val)then return unpack(val) end
	end
	return nil
end

local function doSkinChangeHook(p)
	if(p.lastskin == nil)then
		p.lastskin = p.skin
		enableAllHud()
		doSkinHook(p.skin,"SkinEnabled",p)
		return
	end
	if(p.lastskin == p.skin)then return end
	-- Skin Disabled
	doSkinHook(p.lastskin,"SkinDisabled",p)
	p.lastskin = p.skin
	enableAllHud()
	-- Skin enabled
	doSkinHook(p.skin,"SkinEnabled",p)
end

-- Gametype Hooks

local function doGametypeHook(g,h,...)
	if(not hook.g[g])then return end
	if(not hook.g[g][h])then return nil end
	for i,f in ipairs(hook.s[g][h]) do
		local val = {f(...)}
		if(#val)then return unpack(val) end
	end
	return nil
end

local function doGametypeChangeHook()
	if(lastGametype == nil)then
		lastGametype = gametype
		return
	end
	if(lastGametype == gametype)then return end
	-- Skin Disabled
	doGametypeHook(lastGametype,"GametypeDisabled")
	lastGametype = gametype
end

-- Public

---@param h string Name of the hook
---@param f func Function to call
---@param s skincolor_id ID of skin to use
---@return Namespace, for chaining.
function hook.addSkinHook(h,f,s)
	hook.s[s] = $ or {}
	hook.s[s][h] = $ or {}
	table.insert(hook.s[s][h],f)
	return hook
end

---@param h string Name of the hook
---@param f func Function to call
---@param g gametype_id ID of gametype to use
---@return Namespace, for chaining.
function hook.addGametypeHook(h,f,g)
	hook.g[g] = $ or {}
	hook.g[g][h] = $ or {}
	table.insert(hook.g[g][h],f)
	return hook
end

function hook.isPlayerValid(p,checkmo)
	if(not p)then return false end
	if(not p.valid)then return false end
	if(checkmo and (not p.mo or not p.mo.valid))then return false end
	return true
end

-- Hooks

-- Interaction Hooks

addHook("ShouldDamage", function(mo,...)
	local p = mo.player
	-- Favor gametype over player hooks
	local gtret = doGametypeHook(gametype,"ShouldDamage",p,mo,...)
	if(gtret)then return gtret end
	return doSkinHook(p.skin,"ShouldDamage",p,mo,...)
end,MT_PLAYER)

---@Hook PlayerDamage
---@Description Called when a player is damaged by a mobj.
addHook("MobjDamage", function(mo,...)
	local p = mo.player
	-- Favor gametype over player hooks
	local gtret = doGametypeHook(gametype,"PlayerDamage",p,mo,...)
	if(gtret)then return gtret end
	return doSkinHook(p.skin,"PlayerDamage",p,mo,...)
end,MT_PLAYER)

---@Hook PlayerAttack
---@Description Called when a player is damaging a mobj.
addHook("MobjDamage", function(mo,i,s,...)
	if(not s)then return end
	if(s.type ~= MT_PLAYER)then return end
	local p = s.player
	if(not hook.isPlayerValid(p,true))then return end
	-- Favor gametype over player hooks
	local gtret = doGametypeHook(gametype,"PlayerAttack",p,mo,...)
	if(gtret)then return gtret end
	return doSkinHook(p.skin,"PlayerAttack",p,mo,i,s,...)
end)

---@Hook PlayerDeath
---@Description Called when a player has died.
addHook("MobjDeath",function(mo,...)
	local p = mo.player
	-- Favor gametype over player hooks
	local gtret = doGametypeHook(gametype,"PlayerDeath",p,mo,...)
	if(gtret)then return gtret end
	return doSkinHook(p.skin,"PlayerDeath",p,mo,...)
end,MT_PLAYER)

-- ThinkFrame-related Hooks

addHook("PreThinkFrame", do
	-- Active player check (yes it's every frame...)
	hook.activePlayers = 0
	for p in players.iterate do
		if(p.valid and not p.spectator)then
			hook.activePlayers = $+1
		end
	end
	-- Gametype
	doGametypeHook(gametype,"PreThinkFrame")
	-- Players
	for p in players.iterate do
		if(not hook.isPlayerValid(p,true))then continue end
		doSkinChangeHook(p)
		if(p.deadtimer or p.spectator)then continue end
		doGametypeHook(gametype,"PreThinkFrame",p)
		doSkinHook(p.skin,"PreThinkFrame",p)
	end
end)

addHook("ThinkFrame",do doGametypeHook(gametype,"ThinkFrame") end)

addHook("PostThinkFrame",do
	-- Gametype
	doGametypeHook(gametype,"PostThinkFrame")
	-- Players
	for p in players.iterate do
		if(not hook.isPlayerValid(p,true))then continue end
		if(p.deadtimer or p.spectator)then continue end
		doGametypeHook(gametype,"PostThinkFrame",p)
		doSkinHook(p.skin,"PostThinkFrame",p)
	end
end)

addHook("PlayerThink", function(p)
	if(not hook.isPlayerValid(p,true))then return end
	if(p.deadtimer or p.spectator)then return end
	doGametypeHook(gametype,"PlayerThink",p)
	doSkinHook(p.skin,"PlayerThink",p)
end)

-- Map Hooks

addHook("MapLoad",function(...)
	hook.activePlayers = 0
	for p in players.iterate do
		if(p.valid and not p.spectator)then
			hook.activePlayers = $+1
		end
	end
	doGametypeHook(gametype,"MapLoad",...)
end)

addHook("MapChange",function(...)
	doGametypeChangeHook()
	doGametypeHook(gametype,"MapChange",...)
end)

-- Intermission

addHook("IntermissionThinker", function(...) return doGametypeHook(gametype,"IntermissionThinker",...) end)
addHook("VoteThinker", function(...) return doGametypeHook(gametype,"VoteThinker",...) end)

-- Other Player Hooks

addHook("PlayerCmd", function(p,...)
	if(not hook.isPlayerValid(p,true))then return end
	return doSkinHook(p.skin,"PlayerCmd",p,...)
end)
addHook("PlayerJoin", function(...)		
	return doGametypeHook(gametype,"PlayerJoin",...)	
end)
addHook("PlayerQuit", function(p,...)
	return doGametypeHook(gametype,"PlayerQuit",p,...)	
end)
addHook("TeamSwitch", function(...)			return doGametypeHook(gametype,"TeamSwitch",...)	end)
addHook("PlayerSpawn", function(p)			return doSkinHook(p.skin,"PlayerSpawn",p)		end)
addHook("PlayerCanDamage", function(p,mo)	return doSkinHook(p.skin,"PlayerCanDamage",p,mo)	end)
addHook("ShieldSpawn", function(p)			return doSkinHook(p.skin,"ShieldSpawn",p)		end)
addHook("AbilitySpecial",function(p)		return doSkinHook(p.skin,"AbilitySpecial",p)		end)
addHook("JumpSpecial",function(p)			return doSkinHook(p.skin,"JumpSpecial",p)		end)
addHook("JumpSpinSpecial",function(p)		return doSkinHook(p.skin,"JumpSpinSpecial",p)	end)
addHook("SpinSpecial",function(p)			return doSkinHook(p.skin,"SpinSpecial",p)		end)
addHook("ShieldSpecial",function(p)			return doSkinHook(p.skin,"ShieldSpecial",p)		end)

-- Make global
rawset(_G,"chrishooks_srb2_api",true)
rawset(_G,"chrisHooks",hook)